# import tomli
# import importlib.resources

# def get_version():
#     with importlib.resources.open_text("tace", "pyproject.toml") as f:
#         pyproject = tomli.load(f)
#     return pyproject["project"]["version"]

# __version__ = get_version()


__version__ = "0.0.8"

